/*****************************************************************************
 * Copyright (c) 2020, rrcc FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "FastCall.h"
#include "Stack.h"
#include "Call.h"
#include "../Copy.h"
#include "../Write.h"
#include "../CalcImm.h"

namespace Instruction{

    std::vector<uint8_t> FastCall::toBytes() const {
        std::vector<uint8_t> v;
        for (auto &i : instructions) {
            i->setAddress(getAddress() + v.size());
            auto b = i->toBytes();
            v.insert(v.end(), b.begin(), b.end());
        }
        return v;
    }

    FastCall::FastCall(uint32_t address, const std::vector<Token> &data) {
        initData(data);
        instructions.push_back(std::make_unique<Call>(address));
        takedownData(data);

        comment = "FastCall framing start, " + std::to_string(data.size()) + " arguments";
    }

    FastCall::FastCall(const AddressWithOffset &address, const std::vector<Token> &data) {
        initData(data);
        instructions.push_back(std::make_unique<Call>(address));
        takedownData(data);

        comment = "FastCall framing start, " + std::to_string(data.size()) + " arguments";
    }

    void FastCall::takedownData(const std::vector<Token> &data) {
        instructions.push_back(std::make_unique<Copy>(AddressWithOffset{Stack::frameOffsetRegisterAddress, 0}, AddressWithOffset{1, Stack::stackOffsetRegister}, 1));
        instructions.push_back(std::make_unique<Stack::PopImm>(1 + (data.size() > 2 ? data.size() - 2 : 0)));
    }

    void FastCall::initData(const std::vector<Token> &data) {

        std::vector<std::pair<Stack::PushGroup::Kind, std::pair<uint32_t, AddressWithOffset>>> groupData;

        uint32_t offset = 0;
        for(auto& t : data){
            if(offset >= 2){
                break;
            }else{
                if(t.getType() == Token::Type::RegisterLocation){
                    instructions.push_back(std::make_unique<Copy>(AddressWithOffset{(uint32_t)MgmtRegisters::std_PARAMETER_0 + offset, 0}, AddressWithOffset{t.getNumericValue(), 0}, 1)); //TODO: allow offsets
                }else{
                    instructions.push_back(std::make_unique<Write>(AddressWithOffset{(uint32_t)MgmtRegisters::std_PARAMETER_0 + offset, 0}, std::vector<uint32_t>{t.getNumericValue()}));
                }
            }
            ++offset;
        }

        if(data.size() > 2){

            for(uint32_t i = 2; i < data.size(); ++i){
                auto& t = data.at(i);
                if(t.getType() == Token::Type::RegisterLocation){
                    groupData.push_back({Stack::PushGroup::Kind::Address, {1,AddressWithOffset{t.getNumericValue(), 0}}}); //TODO: allow offsets
                }else{
                    groupData.push_back({Stack::PushGroup::Kind::Immediate, {t.getNumericValue(),AddressWithOffset{0, 0}}});
                }
            }

            /*
            // This uses single push per operation, reverse, left to right
            for(uint32_t i = data.size() - 1; i >= 2; --i){
                auto& t = data.at(i);
                if(t.getType() == Token::Type::RegisterLocation){
                    instructions.push_back(std::make_unique<Stack::Push>(AddressWithOffset{t.getNumericValue(), 0})); //TODO: allow offsets
                }else{
                    instructions.push_back(std::make_unique<Stack::PushImm>(t.getNumericValue()));
                }
            }
            */
        }

        groupData.push_back({Stack::PushGroup::Kind::Address, {1, AddressWithOffset{Stack::frameOffsetRegisterAddress, 0}}});

        instructions.push_back(std::make_unique<Stack::PushGroup>(groupData));

        //instructions.push_back(std::make_unique<Stack::Push>(AddressWithOffset{Stack::frameOffsetRegisterAddress, 0}));
        instructions.push_back(std::make_unique<CalcImm>(Calc::Operation::SUB, 0, Stack::frameOffsetRegisterAddress, Stack::stackOffsetRegisterAddress, 1));
    }

    std::unique_ptr<Instruction> FastCall::fromTokens(const Instruction::parseOperatorResult &op, const std::vector<Token> &tokens) {
        if (tokens.at(0).getType() == Token::Type::CodeLocation || tokens.at(0).getType() == Token::Type::CodeLabel) {
            return std::make_unique<FastCall>(tokens.at(0).getNumericValue(), std::vector<Token>(tokens.begin() + 1, tokens.end()));
        } else {
            return std::make_unique<FastCall>(AddressWithOffset{tokens.at(0).getNumericValue(), op.offset1}, std::vector<Token>(tokens.begin() + 1, tokens.end()));
        }
    }
}